@@ -53,7 +53,6 @@ def iterate(files,banks=None,step=100):
     f = hipochain(files,banks,step=step)
     return f
 
-
 def create(filename):
     """
     Parameters
@@ -113,6 +112,8 @@ class hipofile:
         Extension for buffer file
     buffname : string
         Name of buffer file
+    banklist : dict
+        Dictionary to hipopybind.Bank objects
     """
 
     # Methods
@@ -169,6 +170,7 @@ class hipofile:
         self.dtypes     = {}
         self.buffext    = "~"
         self.buffname   = None
+        self.banklist   = {}
         
     def open(self):
         """
@@ -187,11 +189,6 @@ class hipofile:
             self.reader.open(self.filename)
             self.reader.readDictionary(self.dictionary)
 
-            # Read all existing banks
-            for schema in self.dictionary.getSchemaList():
-                bank = hipopybind.Bank(self.dictionary.getSchema(schema))
-                self.event.getStructure(bank)
-
             # Set group number to highest so far
             self.group = 0
             for schema in self.dictionary.getSchemaList():
@@ -207,6 +204,12 @@ class hipofile:
         elif self.mode=="a" and self.buffname is not None:
             # Now open with writer after adding schema to write
             self.writer.open(self.buffname)
+
+        # Create banks now for speed
+        _dictionary = self.dictionary if self.mode=='r' else self.writer.getDictionary()
+        for schema in _dictionary.getSchemaList():
+            self.banklist[schema] = hipopybind.Bank(_dictionary.getSchema(schema))
+            self.event.getStructure(self.banklist[schema]) #NOTE: IMPORTANT! Necessary before reading banks into event with addStructure.
             
     def flush(self):
         """
@@ -327,8 +330,7 @@ class hipofile:
         Write all existing banks to event for appending to file.
         """
         for schema in self.writer.getDictionary().getSchemaList():
-            bank = hipopybind.Bank(self.writer.getDictionary().getSchema(schema))
-            self.event.getStructure(bank)
+            bank = self.banklist[schema]
             self.event.addStructure(bank)
 
     def writeBank(self, name, names, data, dtypes="D"):
@@ -349,31 +351,26 @@ class hipofile:
         -----------
         Fill an event bank with data and write to buffer.
         """
-        schema = self.writer.getDictionary().getSchema(name)
         rows   = np.shape(data)[-1]
-        bank   = hipopybind.Bank(schema,rows)
+        bank   = self.banklist[name]
+        bank.reset()
+        bank.setRows(rows)
 
         # Add data to bank
         for idx, entry in enumerate(names):
             dtype = dtypes if len(dtypes)==1 else dtypes[idx]
             if dtype=="D":
-                for i in range(rows):
-                    bank.putDouble(entry,i,data[idx,i])
-            elif dtype=="F":
-                for i in range(rows):
-                    bank.putFloat(entry,i,data[idx,i])
+                bank.putDoubles(entry,data[idx])
             elif dtype=="I":
-                for i in range(rows):
-                    bank.putInt(entry,i,data[idx,i])
+                bank.putInts(entry,data[idx])
+            elif dtype=="F":
+                bank.putFloats(entry,data[idx])
             elif dtype=="B":
-                for i in range(rows):
-                    bank.putByte(entry,i,data[idx,i])
+                bank.putBytes(entry,data[idx])
             elif dtype=="S":
-                for i in range(rows):
-                    bank.putShort(entry,i,data[idx,i])
+                bank.putShorts(entry,data[idx])
             elif dtype=="L":
-                for i in range(rows):
-                    bank.putLong(entry,i,data[idx,i])
+                bank.putLongs(entry,data[idx])
             else:
                 raise TypeError
 
@@ -397,7 +394,7 @@ class hipofile:
 
         Description
         -----------
-        Mimics uproot newTree function.
+        Create a new bank to which to add data.  Mimics uproot newtree function.
         """
         group = self.group if group is None else group
         item  = self.item if item is None else item
@@ -413,8 +410,7 @@ class hipofile:
 
         Description
         -----------
-        Mimics uproot extend function. NOTE: dtype argument fixed until I figure out how to pass
-        different types to C wrapper.
+        Add batched data to banks in a file.  Mimics uproot extend function.
         """
         keys = list(datadict.keys())
         nEvents = len(datadict[keys[0]])
@@ -424,8 +420,9 @@ class hipofile:
             for event in range(nEvents):
                 for bank in datadict: # This requires datadict shape to be (nEvents,nNames,nRows)
                     self.writeBank(bank,list(self.dtypes[bank].keys()),datadict[bank][event],dtypes=list(self.dtypes[bank].values()))
-                self.addEvent()
-            self.writeEvent()
+                self.writer.addEvent(self.event)
+                self.event.reset()
+            self.writer.flush()
 
         # Append mode routine
         elif self.mode == "a":
@@ -435,8 +432,9 @@ class hipofile:
                     break
                 for bank in datadict: # This requires datadict shape to be (nEvents,nNames,nRows)
                     self.writeBank(bank,self.dtypes[bank].keys(),datadict[bank][event],dtypes=list(self.dtypes[bank].values()))
-                self.addEvent()
-            self.writeEvent()
+                self.writer.addEvent(self.event)
+                self.event.reset()
+            self.writer.flush()
 
     def update(self,datadict):
         """
@@ -448,15 +446,15 @@ class hipofile:
         Description
         -----------
         Append one set of event banks at a time and do not progress to the 
-        next event automatically. NOTE: dtype argument fixed until I figure out how to pass
-        different types to C wrapper.
+        next event automatically.
         """
         # Append mode routine
         if self.mode == "a":
             for bank in datadict: # This requires datadict shape to be (nNames,nRows)
                 self.writeBank(bank,self.dtypes[bank].keys(),datadict[bank],dtypes=list(self.dtypes[bank].values()))
-            self.addEvent()
-            self.writeEvent()
+            self.writer.addEvent(self.event)
+            self.event.reset()
+            self.writer.flush()
 
     def write(self):
         """
@@ -517,7 +515,6 @@ class hipofile:
             bank = hipopybind.Bank(self.dictionary.getSchema(schema))
             self.event.getStructure(bank)
 
-
     def readBank(self,bankName,verbose=False):
         """
         Parameters
@@ -548,7 +545,6 @@ class hipofile:
         self.group = group
         return self.group
 
-
     def getEntries(self,bankName):
         """
         Parameters
@@ -570,16 +566,23 @@ class hipofile:
 
         Returns
         -------
-        data : ctypes.c_char_p array
-            ctypes c_char_p array containing the bank entries
+        data : dict
+            Dictionary of bank entry names to types (represented as strings)
 
         Description
         -----------
         Get a list of the entry names from the data table in the current event's bank.
         """
-        bankdict = self.dictionary.getSchema(bankName).getSchemaString()
-        bankdict = bankdict.split("}{")[1][:-1]
-        bankdict = { entry.split("/")[0]:entry.split("/")[1] for entry in bankdict.split(",")}
+        bankdict = None
+        try:
+            bankdict = self.dictionary.getSchema(bankName).getSchemaString()
+            bankdict = bankdict.split("}{")[1][:-1]
+            bankdict = { entry.split("/")[0]:entry.split("/")[1] for entry in bankdict.split(",")}
+        except IndexError:
+            print("hipopy.hipopy.hipofile.getNamesAndTypes schemaString unreadable")
+            print("bankName = ",bankName)
+            print("schemaString = ",self.dictionary.getSchema(bankName).getSchemaString())
+            return None
         return bankdict
     
     def getNames(self,bankName):
@@ -590,8 +593,8 @@ class hipofile:
 
         Returns
         -------
-        data : ctypes c_char_p array
-            ctypes c_char_p array containing the bank entries
+        data : list
+            List of entry names in bank
 
         Description
         -----------
@@ -609,7 +612,8 @@ class hipofile:
 
         Returns
         -------
-        data : ctypes c_char_p array containing the bank entries
+        data : list
+            List of entry types (represented as strings) in bank
 
         Description
         -----------
@@ -631,8 +635,7 @@ class hipofile:
         Get number of rows in bank.  Make sure you read bank first 
         with readBank(bankName) method above.
         """
-        bank = hipopybind.Bank(self.dictionary.getSchema(bankName))
-        self.event.getStructure(bank)
+        bank = self.banklist[bankName]
         return bank.getRows()
 
     def getInts(self,bankName,item):
@@ -641,21 +644,19 @@ class hipofile:
         ----------
         bankName : string, required
         item : string, required
-            Column name you wish to read in bank
+            Column name to read in bank
 
         Returns
         -------
-        data : ctypes int array
-            ctypes int array containing the bank entries
+        data : list
+            List of ints from bank entry
 
         Description
         -----------
         Get a column of ints from the data table in the current event's bank.
         """
-        bank = hipopybind.Bank(self.dictionary.getSchema(bankName))
-        self.event.getStructure(bank)
-        bankRows = bank.getRows()
-        data = [bank.getInt(item,i) for i in range(bankRows)]
+        bank = self.banklist[bankName]
+        data = bank.getInts(item)
         return data
 
     def getFloats(self,bankName,item):
@@ -664,21 +665,19 @@ class hipofile:
         ----------
         bankName : string, required
         item : string, required
-            Column name you wish to read in bank
+            Column name to read in bank
 
         Returns
         -------
-        data : cyptes float array
-            ctypes float array containing the bank entries
+        data : list
+            List of floats from bank entry
 
         Description
         -----------
         Get a column of floats from the data table in the current event's bank.
         """
-        bank = hipopybind.Bank(self.dictionary.getSchema(bankName))
-        self.event.getStructure(bank)
-        bankRows = bank.getRows()
-        data = [bank.getFloat(item,i) for i in range(bankRows)]
+        bank = self.banklist[bankName]
+        data = bank.getFloats(item)
         return data
 
     def getDoubles(self,bankName,item):
@@ -687,96 +686,84 @@ class hipofile:
         ----------
         bankName : string, required
         item : string, required
-            Column name you wish to read in bank
+            Column name to read in bank
 
         Returns
         -------
-        data : ctypes double array
-            ctypes double array containing the bank entries
+        data : list
+            List of doubles from bank entry
 
         Description
         -----------
         Get a column of doubles from the data table in the current event's bank.
         """
-        bank = hipopybind.Bank(self.dictionary.getSchema(bankName))
-        self.event.getStructure(bank)
-        bankRows = bank.getRows()
-        data = [bank.getDouble(item,i) for i in range(bankRows)]
+        bank = self.banklist[bankName]
+        data = bank.getDoubles(item)
         return data
 
-
     def getShorts(self,bankName,item):
         """
         Parameters
         ----------
         bankName : string, required
         item : string, required
-            Column name you wish to read in bank
+            Column name to read in bank
 
         Returns
         -------
-        data : ctypes short array
-            ctypes short array containing the bank entries
+        data : list
+            List of shorts from bank entry
 
         Description
         -----------
         Get a column of shorts from the data table in the current event's bank.
         """
-        bank = hipopybind.Bank(self.dictionary.getSchema(bankName))
-        self.event.getStructure(bank)
-        bankRows = bank.getRows()
-        data = [bank.getShort(item,i) for i in range(bankRows)]
+        bank = self.banklist[bankName]
+        data = bank.getShorts(item)
         return data
 
-
     def getLongs(self,bankName,item):
         """
         Parameters
         ----------
         bankName : string, required
         item : string, required
-            Column name you wish to read in bank
+            Column name to read in bank
 
         Returns
         -------
-        data : ctypes long array
-            ctypes long array containing the bank entries
+        data : list
+            List of longs from bank entry
 
         Description
         -----------
         Get a column of longs from the data table in the current event's bank.
         """
-        bank = hipopybind.Bank(self.dictionary.getSchema(bankName))
-        self.event.getStructure(bank)
-        bankRows = bank.getRows()
-        data = [bank.getLong(item,i) for i in range(bankRows)]
+        bank = self.banklist[bankName]
+        data = bank.getLongs(item)
         return data
 
-
     def getBytes(self,bankName,item):
         """
         Parameters
         ----------
         bankName : string, required
         item : string, required
-            Column name you wish to read in bank
+            Column name to read in bank
 
         Returns
         -------
-        data : ctypes c_int array
-            ctypes c_int array containing the bank entries
+        data : list
+            List of bytes from bank entry
 
         Description
         -----------
         Get a column of bytes from the data table in the current event's bank.
         """
-        bank = hipopybind.Bank(self.dictionary.getSchema(bankName))
-        self.event.getStructure(bank)
-        bankRows = bank.getRows()
-        data = [bank.getByte(item,i) for i in range(bankRows)]
+        bank = self.banklist[bankName]
+        data = bank.getBytes(item)
         return data
 
-
     def __iter__(self):
         return hipofileIterator(self)
 
@@ -813,15 +800,15 @@ class hipofileIterator:
             for bank in self.banks:
                 for item in self.items[bank]:
                     data = []
-                    if   self.items[bank][item]=="F": data = self.hipofile.getFloats(bank,item)
+                    if   self.items[bank][item]=="D": data = self.hipofile.getDoubles(bank,item)
                     elif self.items[bank][item]=="I": data = self.hipofile.getInts(bank,item)
-                    elif self.items[bank][item]=="D": data = self.hipofile.getDoubles(bank,item)
+                    elif self.items[bank][item]=="F": data = self.hipofile.getFloats(bank,item)
                     elif self.items[bank][item]=="L": data = self.hipofile.getLongs(bank,item)
                     elif self.items[bank][item]=="S": data = self.hipofile.getShorts(bank,item)
                     elif self.items[bank][item]=="B": data = self.hipofile.getBytes(bank,item)
 
                     # Add bank data to event dictionary
-                    event[bank+"_"+item] = [np.array(data)]
+                    event[bank+"_"+item] = [data]
 
             return event
         raise StopIteration
@@ -947,16 +934,16 @@ class hipochainIterator:
                 for bank in self.chain.banks:
                     for item in self.items[bank]:
                         data = []
-                        if   self.items[bank][item]=="F": data = self.file.getFloats(bank,item)
+                        if   self.items[bank][item]=="D": data = self.file.getDoubles(bank,item)
                         elif self.items[bank][item]=="I": data = self.file.getInts(bank,item)
-                        elif self.items[bank][item]=="D": data = self.file.getDoubles(bank,item)
+                        elif self.items[bank][item]=="F": data = self.file.getFloats(bank,item)
                         elif self.items[bank][item]=="L": data = self.file.getLongs(bank,item)
                         elif self.items[bank][item]=="S": data = self.file.getShorts(bank,item)
                         elif self.items[bank][item]=="B": data = self.file.getBytes(bank,item)
 
                         # Add bank data to batch dictionary
-                        if not bank+"_"+item in self.dict.keys() : self.dict[bank+"_"+item] = [np.array(data)]
-                        else: self.dict[bank+"_"+item].append(np.array(data))               
+                        if not bank+"_"+item in self.dict.keys() : self.dict[bank+"_"+item] = [data]
+                        else: self.dict[bank+"_"+item].append(data)            
 
                 # Check size of output array
                 self.counter += 1
